import { dirname, extname, normalize, resolve, sep, join } from 'path';
import builtins from 'builtin-modules';
import resolveId from 'resolve';
import isModule from 'is-module';
import fs from 'fs';

var ES6_BROWSER_EMPTY = resolve( __dirname, '../src/empty.js' );

// It is important that .mjs occur before .js so that Rollup will interpret npm modules
// which deploy both ESM .mjs and CommonJS .js files as ESM.
var DEFAULT_EXTS = [ '.mjs', '.js', '.json', '.node' ];

var readFileCache = {};
var readFileAsync = function (file) { return new Promise(function (fulfil, reject) { return fs.readFile(file, function (err, contents) { return err ? reject(err) : fulfil(contents); }); }); };
var statAsync = function (file) { return new Promise(function (fulfil, reject) { return fs.stat(file, function (err, contents) { return err ? reject(err) : fulfil(contents); }); }); };

function cachedReadFile (file, cb) {
	if (file in readFileCache === false) {
		readFileCache[file] = readFileAsync(file).catch(function (err) {
			delete readFileCache[file];
			throw err;
		});
	}
	readFileCache[file].then(function (contents) { return cb(null, contents); }, cb);
}

var isFileCache = {};
function cachedIsFile (file, cb) {
	if (file in isFileCache === false) {
		isFileCache[file] = statAsync(file)
			.then(
				function (stat) { return stat.isFile(); },
				function (err) {
					if (err.code == 'ENOENT') { return false; }
					delete isFileCache[file];
					throw err;
				});
	}
	isFileCache[file].then(function (contents) { return cb(null, contents); }, cb);
}

function deprecatedMainField (options, option, mainFields, field) {
	if ( field === void 0 ) field = option;

	if (option in options) {
		if (options[option] === false) {
			return mainFields.filter(function (mainField) { return mainField !== field; });
		} else if (options[option] === true && !mainFields.includes(field)) {
			return mainFields.concat(field);
		}
	}
	return mainFields;
}

var resolveIdAsync = function (file, opts) { return new Promise(function (fulfil, reject) { return resolveId(file, opts, function (err, contents) { return err ? reject(err) : fulfil(contents); }); }); };

// resolve alias helpers
var isAlias = function (file, alias) {
	if (alias === file) {
		return true;
	}
	if (!file.startsWith(alias)) {
		return false;
	}
	return file[alias.length] === '/';
};
var getAlias = function (file, aliases, offset) {
	for (var i = offset, l = aliases.length, o = (void 0); i < l; ++i) {
		if ((o = aliases[i]) && isAlias(file, o[0])) { return [ i, o ]; }
	}
	return null;
};
var localImport = /^[.]{1,2}\//;
var resolveAliases = function (target, aliases, offset) {
	if ( offset === void 0 ) offset = 0;

	if (localImport.test(target)) { return null; }
	var tuple = getAlias(target, aliases, offset);
	if (tuple === null) { return null; }
	var cursor = tuple[0];
	var tuple_1 = tuple[1];
	var alias = tuple_1[0];
	var p = tuple_1[1]; // eslint-disable-line prefer-const
	p = join(p, target.substr(alias.length));
	return resolveAliases(p, aliases, ++cursor) || p;
};

var FALSE = {};

function nodeResolve ( options ) {
	if ( options === void 0 ) options = {};

	if ('mainFields' in options && ('module' in options || 'main' in options || 'jsnext' in options)) {
		throw new Error("node-resolve: do not use deprecated 'module', 'main', 'jsnext' options with 'mainFields'");
	}

	var mainFields = options.mainFields || [];
	var fields = { 'browser': 0, 'module': 0, 'jsnext': 'jsnext:main', 'main': 0 };
	var keys = Object.keys(options).filter(function (k) { return fields.hasOwnProperty(k); });

	// build mainFields by plugins options (keep options order)
	mainFields = keys.reduce(function (list, k) { return deprecatedMainField(options, k, list, fields[k] || k); }, mainFields);

	// set defaults mainFields if none options
	[ 'module', 'main' ].forEach(function (k) {
		if (!mainFields.includes(k) && options[k] !== false) { mainFields.push(k); }
	});

	var isPreferBuiltinsSet = options.preferBuiltins === true || options.preferBuiltins === false;
	var preferBuiltins = isPreferBuiltinsSet ? options.preferBuiltins : true;
	var customResolveOptions = options.customResolveOptions || {};
	var jail = options.jail;
	var only = Array.isArray(options.only)
		? options.only.map(function (o) { return o instanceof RegExp
			? o
			: new RegExp('^' + String(o).replace(/[\\^$*+?.()|[\]{}]/g, '\\$&') + '$'); }
		)
		: null;
	var browserMapCache = {};

	if ( options.skip ) {
		throw new Error( 'options.skip is no longer supported — you should use the main Rollup `external` option instead' );
	}

	if ( !mainFields.length ) {
		throw new Error( "Please ensure at least one 'mainFields' value is specified" );
	}

	// { k/v, ... } => [ [k, v], ... ]
	var kvs = options.alias || {};
	var aliases = Object.keys(kvs).reduce(function (p, k) { return (p.push([ k, kvs[k] ]), p); }, []); // eslint-disable-line no-sequences

	var preserveSymlinks;

	return {
		name: 'node-resolve',

		options: function options ( options$1 ) {
			preserveSymlinks = options$1.preserveSymlinks;
		},

		generateBundle: function generateBundle () {
			isFileCache = {};
			readFileCache = {};
		},

		resolveId: function resolveId$$1 ( importee, importer ) {
			var this$1 = this;

			if ( /\0/.test( importee ) ) { return null; } // ignore IDs with null character, these belong to other plugins

			var basedir = importer ? dirname( importer ) : process.cwd();

			// check aliases first
			if (aliases.length) {
				var alias = resolveAliases(importee, aliases);
				if (alias) {
					importee = alias;
				}
			}

			// https://github.com/defunctzombie/package-browser-field-spec
			if (mainFields.includes('browser') && browserMapCache[importer]) {
				var resolvedImportee = resolve( basedir, importee );
				var browser = browserMapCache[importer];
				if (browser[importee] === false || browser[resolvedImportee] === false) {
					return ES6_BROWSER_EMPTY;
				}
				if (browser[importee] || browser[resolvedImportee] || browser[resolvedImportee + '.js'] || browser[resolvedImportee + '.json']) {
					importee = browser[importee] || browser[resolvedImportee] || browser[resolvedImportee + '.js'] || browser[resolvedImportee + '.json'];
				}
			}


			var parts = importee.split( /[/\\]/ );
			var id = parts.shift();

			if ( id[0] === '@' && parts.length ) {
				// scoped packages
				id += "/" + (parts.shift());
			} else if ( id[0] === '.' ) {
				// an import relative to the parent dir of the importer
				id = resolve( basedir, importee );
			}

			if (only && !only.some(function (pattern) { return pattern.test(id); })) { return null; }

			var disregardResult = false;
			var packageBrowserField = false;
			var extensions = options.extensions || DEFAULT_EXTS;

			var resolveOptions = {
				basedir: basedir,
				packageFilter: function packageFilter ( pkg, pkgPath ) {
					var pkgRoot = dirname( pkgPath );
					if (mainFields.indexOf('browser') !== -1 && typeof pkg[ 'browser' ] === 'object') {
						packageBrowserField = Object.keys(pkg[ 'browser' ]).reduce(function (browser, key) {
							var resolved = pkg[ 'browser' ][ key ] === false ? false : resolve( pkgRoot, pkg[ 'browser' ][ key ] );
							browser[ key ] = resolved;
							if ( key[0] === '.' ) {
								var absoluteKey = resolve( pkgRoot, key );
								browser[ absoluteKey ] = resolved;
								if ( !extname(key) ) {
									extensions.reduce( function ( browser, ext ) {
										browser[ absoluteKey + ext ] = browser[ key ];
										return browser;
									}, browser );
								}
							}
							return browser;
						}, {});
					}

					var overriddenMain = false;
					for ( var i in mainFields ) {
						var field = mainFields[i];
						if ( field !== 'main' && typeof pkg[ field ] === 'string' ) {
							pkg[ 'main' ] = pkg[ field ];
							overriddenMain = true;
							break;
						}
					}
					if ( overriddenMain === false && mainFields.indexOf( 'main' ) === -1 ) {
						disregardResult = true;
					}
					return pkg;
				},
				readFile: cachedReadFile,
				isFile: cachedIsFile,
				extensions: extensions
			};

			if (preserveSymlinks !== undefined) {
				resolveOptions.preserveSymlinks = preserveSymlinks;
			}

			return resolveIdAsync(
				importee,
				Object.assign( resolveOptions, customResolveOptions )
			)
				.catch(function () { return FALSE; })
				.then(function (resolved) {
					if (resolved !== FALSE && mainFields.includes('browser') && packageBrowserField) {
						var l;
						var hasLink = packageBrowserField.hasOwnProperty(resolved);
						if (hasLink) {
							l = packageBrowserField[ resolved ];
						}
						browserMapCache[l || resolved] = packageBrowserField;
						if (hasLink) {
							resolved = l;
						}
						if (resolved === false) {
							return ES6_BROWSER_EMPTY;
						}
					}

					if (resolved === FALSE) {
						return null;
					}

					if ( !disregardResult ) {
						if ( !preserveSymlinks && resolved && fs.existsSync( resolved ) ) {
							resolved = fs.realpathSync( resolved );
						}

						if ( ~builtins.indexOf( resolved ) ) {
							return null;
						} else if ( ~builtins.indexOf( importee ) && preferBuiltins ) {
							if ( !isPreferBuiltinsSet ) {
								this$1.warn(
									"preferring built-in module '" + importee + "' over local alternative " +
									"at '" + resolved + "', pass 'preferBuiltins: false' to disable this " +
									"behavior or 'preferBuiltins: true' to disable this warning"
								);
							}
							return null;
						} else if ( jail && resolved.indexOf( normalize( jail.trim( sep ) ) ) !== 0 ) {
							return null;
						}
					}

					if ( resolved && options.modulesOnly ) {
						return readFileAsync( resolved, 'utf-8').then(function (code) { return isModule( code ) ? resolved : null; });
					} else {
						return resolved;
					}
				});
		}
	};
}

export default nodeResolve;
